#!/usr/bin/python

"""
C55Battleship - A battleship game by Perttu "celeron55" Ahola.

Made as the 'Ohjelmoinnin alkeet / perusteet' 2009 final assignment.

- The AI is "fair"; it doesn't use information that a real player wouldn't know.
- Save game files use an XML format.
- Scoring wasn't implemented because the game type doesn't have trivial support for them.
"""

import Tkinter
import tkMessageBox
import tkFileDialog
from xml.dom.minidom import Document
from xml.dom import minidom

from battleshipbase import *

imagepath = "images"

images_to_load = [
"water","waterunknown","explosion","splash",
"patrol","submarine","destroyer","battleship","carrier",
"patrol_90","submarine_90","destroyer_90","battleship_90","carrier_90",
"patrol_d","submarine_d","destroyer_d","battleship_d","carrier_d",
"patrol_90_d","submarine_90_d","destroyer_90_d","battleship_90_d","carrier_90_d"
]

MODE_NORMAL = 0
MODE_MANUAL_PLACEMENT = 1

class Game:
	def __init__(self, w, h, images):
		"""
		Initialize Game
		
		w, h: width, height of map
		images: a dictionary of the preloaded images
		"""
		self.mode = MODE_NORMAL
		self.ships_to_add = [] # for use of the manual placement mode
		self.ship_horizontal = False # for use of the manual placement mode
		self.player_map = Map(w, h, images)
		self.computer_map = Map(w, h, images)
		self.game_over = True
		
	def load(self, filename):
		global info_text
		file = open(filename, 'r')
		doc = minidom.parse(file)
		file.close()
		
		top = doc.getElementsByTagName("C55BattleshipGame")[0]

		self.mode = int(top.getAttribute("mode"))
		self.game_over = bool(int( top.getAttribute("game_over") ))
		self.ship_horizontal = bool(int( top.getAttribute("ship_horizontal") ))
		info_text.set(top.getAttribute("info_text"))

		tmp = top.getElementsByTagName("ships_to_add")[0]
		textnode = tmp.childNodes[0]
		del self.ships_to_add[0:]
		for i in textnode.toxml().split(','):
			i = i.strip(' \t\n\r')
			if i!="":
				self.ships_to_add.append(i)


		tmp = top.getElementsByTagName("PlayerMap")[0]
		self.player_map.read_xml(tmp)
		tmp = top.getElementsByTagName("ComputerMap")[0]
		self.computer_map.read_xml(tmp)

	def save(self, filename):
		global info_text
		doc = Document()

		top = doc.createElement("C55BattleshipGame")
		
		top.setAttribute("mode", str(self.mode))
		top.setAttribute("game_over", str(int(self.game_over)))
		top.setAttribute("ship_horizontal", str(int(self.ship_horizontal)))
		top.setAttribute("info_text", info_text.get())

		doc.appendChild(top)

		list_string = ""
		i = 0
		for name in self.ships_to_add:
			list_string = list_string + name
			if i != len(self.ships_to_add)-1:
				list_string = list_string + ","
			i = i + 1
		tmp = doc.createElement("ships_to_add")
		textnode = doc.createTextNode(list_string)
		tmp.appendChild(textnode)
		top.appendChild(tmp)

		tmp = doc.createElement("PlayerMap")
		self.player_map.write_xml(doc, tmp)
		top.appendChild(tmp)

		tmp = doc.createElement("ComputerMap")
		self.computer_map.write_xml(doc, tmp)
		top.appendChild(tmp)

		file = open(filename, 'w')
		doc.writexml(file, newl="\n", addindent=" ")
		file.close()


# Functions to be called from the GUI

def new_game(force_autoplace=False):
	global game, info_text, canvas_player, label_right
	
	game.player_map.clear()
	game.computer_map.clear()
	game.game_over = False

	game.computer_map.add_ships_randomly() # Add the opponent's ships randomly
	
	manual_placement = False
	if not force_autoplace:
		manual_placement = tkMessageBox.askyesno("Battleship", "Place ships yourself?")

	if manual_placement:
		del game.ships_to_add[0:]
		game.ships_to_add = get_ship_add_list()
		game.mode = MODE_MANUAL_PLACEMENT
		info_text.set("Adding "+game.ships_to_add[0])
		label_right.config(text="Lclick = place ship, Rclick = toggle vert/horiz")
	else:
		# Automatic placement
		game.player_map.add_ships_randomly()
		game.mode = MODE_NORMAL
		if not force_autoplace:
			info_text.set("That's the spirit! Shoot 'em up!")
		else:
			info_text.set("Greetings from the Outer Space.")

	canvas_computer.delete(Tkinter.ALL)
	game.computer_map.draw(canvas_computer, False)

	canvas_player.delete(Tkinter.ALL)
	game.player_map.draw(canvas_player, True)

def load_game():
	global game, info_text
	filename = tkFileDialog.askopenfilename(initialdir="./",
		title="Select a file to load", filetypes=[("Battleship game files",".bgf")])
	if not filename:
		info_text.set("Loading nothing.")
		return
	game.load(filename)

	canvas_computer.delete(Tkinter.ALL)
	game.computer_map.draw(canvas_computer, False)

	canvas_player.delete(Tkinter.ALL)
	game.player_map.draw(canvas_player, True)

def save_game():
	global game, info_text
	filename = tkFileDialog.asksaveasfilename(initialdir="./", initialfile="battleship_save.bgf",
		title="Select a file name for saving", filetypes=[("Battleship game files",".bgf")])
	if not filename:
		info_text.set("Saving nothing.")
		return
	game.save(filename)

	canvas_computer.delete(Tkinter.ALL)
	game.computer_map.draw(canvas_computer, False)

	canvas_player.delete(Tkinter.ALL)
	game.player_map.draw(canvas_player, True)


def about():
	global __doc__
	tkMessageBox.showinfo("About", __doc__)

def canvas_computer_click(event):
	global game, canvas_computer, canvas_player, info_text
	if game.game_over:
		return
	everything_gone = False
	if game.mode == MODE_NORMAL:
		# The player clicked a position on the computer's map.
		x, y = canvas_to_map(event.x, event.y)
		valid_shot = False
		try:
			everything_gone, message = game.computer_map.shoot((x,y))
			info_text.set(message)
			valid_shot = True
		except ValueError:
			info_text.set("Couldn't shoot here")
			
		canvas_computer.delete(Tkinter.ALL)
		game.computer_map.draw(canvas_computer, False)

		if not valid_shot:
			return
		
		if everything_gone:
			# Every ship was destroyed, the user won the game.
			# Fully draw the computer's map
			canvas_computer.delete(Tkinter.ALL)
			game.computer_map.draw(canvas_computer, True)

			tkMessageBox.showinfo("Game over", "You win!")
			info_text.set("Game over, you won.")
			game.game_over = True
			return

		# Now it's the computer's turn:
		everything_gone, position = game.player_map.shoot_somewhere()

		canvas_player.delete(Tkinter.ALL)
		game.player_map.draw(canvas_player, True)

		if everything_gone:
			# Every ship was destroyed, the computer won the game.
			# Draw the computer's ships
			canvas_computer.delete(Tkinter.ALL)
			game.computer_map.draw(canvas_computer, True)

			tkMessageBox.showinfo("Game over", "The computer wins!")
			info_text.set("Game over, the computer won.")

	elif game.mode == MODE_MANUAL_PLACEMENT:
		pass

def canvas_player_click(event):
	global game, canvas_player, label_right
	if game.game_over:
		return
	if game.mode == MODE_NORMAL:
		pass
	elif game.mode == MODE_MANUAL_PLACEMENT:
		x, y = canvas_to_map(event.x, event.y)
		name = game.ships_to_add[0]
		ship = Ship(images)
		ship.set(x, y, game.ship_horizontal, name)
		try:
			game.player_map.add_ship(ship)
			game.ships_to_add.pop(0)
			if len(game.ships_to_add) == 0:
				info_text.set("All ships added. Start shooting!")
				label_right.config(text="Shoot at the left, curse at the right.")
				game.mode = MODE_NORMAL
			else:
				info_text.set("Adding "+game.ships_to_add[0])
		except ValueError:
			info_text.set("Invalid ship placement")
	canvas_player.delete(Tkinter.ALL)
	game.player_map.draw(canvas_player, True)
	
		
def canvas_player_click3(event):
	global game
	if game.game_over:
		return
	if game.mode == MODE_NORMAL:
		pass
	elif game.mode == MODE_MANUAL_PLACEMENT:
		game.ship_horizontal = not game.ship_horizontal
		pass

if __name__ == "__main__":

	W = 10
	H = 10

	# A dictionary containing images
	images = dict()

	# Initialization

	random.seed()

	game = Game(W, H, images)

	root = Tkinter.Tk()
	root.title("Battleship by celeron55")

	# Load images

	for name in images_to_load:
		filename = imagepath+"/"+name+".gif"
		#print "loading "+filename
		images[name] = Tkinter.PhotoImage(file=filename)

	# Set up all the GUI stuff

	menubar = Tkinter.Menu(root)
	gamemenu = Tkinter.Menu(menubar, tearoff=0)
	gamemenu.add_command(label="New", command=new_game)
	gamemenu.add_command(label="Load", command=load_game)
	gamemenu.add_command(label="Save", command=save_game)
	menubar.add_cascade(label="Game", menu=gamemenu)
	menubar.add_command(label="About", command=about)
	root.config(menu=menubar)

	w = Tkinter.Label(root, text="Computer:").grid(row=0, column=0)
	canvas_computer = Tkinter.Canvas(root, width=W*BLOCK_W, height=H*BLOCK_H)
	canvas_computer.grid(row=1, column=0)
	canvas_computer.bind("<Button-1>", canvas_computer_click)

	w = Tkinter.Label(root, text="You:").grid(row=0, column=1, columnspan=1)
	canvas_player = Tkinter.Canvas(root, width=W*BLOCK_W, height=H*BLOCK_H)
	canvas_player.grid(row=1, column=1, columnspan=1)
	canvas_player.bind("<Button-1>", canvas_player_click)
	canvas_player.bind("<Button-3>", canvas_player_click3)

	label_right = Tkinter.Label(root, text="Shoot at the left, curse at the right.")
	label_right.grid(row=2, column=1)

	info_text = Tkinter.StringVar()
	info_text.set("<info>")
	label_info = Tkinter.Label(root, textvariable=info_text);
	label_info.grid(row=2, column=0)

	# Update the game views

	canvas_computer.delete(Tkinter.ALL)
	game.computer_map.draw(canvas_computer, False)

	canvas_player.delete(Tkinter.ALL)
	game.player_map.draw(canvas_player, True)

	# Now make a new game and start waiting for events

	new_game(True)

	Tkinter.mainloop()

