import Tkinter
import random
import operator

BLOCK_W = 32
BLOCK_H = 32

def canvas_to_map(x, y):
	return (x / BLOCK_W, y / BLOCK_H)

ship_info = {
#   name:    (length, number of ships used, short name)
	'Patrol Boat': (2, 2, "patrol"),
	'Submarine':   (3, 2, "submarine"),
	'Destroyer':   (3, 2, "destroyer"),
	'Battleship':  (4, 1, "battleship"),
	'Aircraft Carrier': (5, 1, "carrier")
}

def get_ship_add_list():
	ships_to_add = []
	for name, (length, count, short) in ship_info.iteritems():
		for i in range(0, count):
			ships_to_add.append(name)
	return ships_to_add

class Ship:
	"""A ship object on the map"""

	def __init__(self, images):
		"""
		Initialize the Ship object.
		
		images: a dictionary of the preloaded images
		"""
		self.x = 0
		self.y = 0
		self.horizontal = True
		self.name = ""
		self.length = 0
		self.hits = 0
		self.images = images
	
	def set(self, x, y, horizontal, name):
		self.x = x
		self.y = y
		self.horizontal = horizontal
		self.name = name
		self.length = ship_info[name][0]
		self.hits = 0
	
	def get_positions(self):
		"""Get a list of the coordinate positions of the ship"""
		x = self.x
		y = self.y
		positions = []
		for i in range(0, self.length):
			positions.append((x,y))
			if self.horizontal:
				x = x + 1
			else:
				y = y + 1
		return positions
	
	def add_hit(self):
		"""Add a hit to the ship. Returns True if ship is destroyed."""
		self.hits = self.hits + 1
		return self.destroyed()
	
	def destroyed(self):
		"""Returns a boolean indicating whether the ship is destroyed or not."""
		if self.hits == self.length:
			return True
		else:
			return False
	
	def draw(self, canvas, x=0, y=0):
		if not self.destroyed():
			if self.horizontal:
				canvas.create_image(x+self.x*BLOCK_W, y+self.y*BLOCK_H, image=self.images[ship_info[self.name][2]], anchor=Tkinter.NW)
			else:
				canvas.create_image(x+self.x*BLOCK_W, y+self.y*BLOCK_H, image=self.images[ship_info[self.name][2]+"_90"], anchor=Tkinter.NW)
		else:
			if self.horizontal:
				canvas.create_image(x+self.x*BLOCK_W, y+self.y*BLOCK_H, image=self.images[ship_info[self.name][2]+"_d"], anchor=Tkinter.NW)
			else:
				canvas.create_image(x+self.x*BLOCK_W, y+self.y*BLOCK_H, image=self.images[ship_info[self.name][2]+"_90_d"], anchor=Tkinter.NW)
	
	def write_xml(self, doc, element):
		element.setAttribute("length", str(self.length))
		element.setAttribute("hits", str(self.hits))
		element.setAttribute("x", str(self.x))
		element.setAttribute("y", str(self.y))
		element.setAttribute("horizontal", str(int(self.horizontal)))
		element.setAttribute("name", self.name)
	
	def read_xml(self, element):
		self.length = int(element.getAttribute("length"))
		self.hits = int(element.getAttribute("hits"))
		self.x = int(element.getAttribute("x"))
		self.y = int(element.getAttribute("y"))
		self.horizontal = bool(int(element.getAttribute("horizontal")))
		self.name = element.getAttribute("name")

class Block:
	"""Contents of a single coordinate position on the map"""

	def __init__(self):
		self.ship_number = -1 # -1: no ship; >=0: The index of the ship list element describing the ship.
		self.has_shot = False # True if the position has been shot at.
	
	def print_(self, show_ships):
		"""Print a single character followed by a space describing the content"""
		if self.has_shot:
			if self.ship_number != -1:
				print 'X',
			else:
				print 'o',
		elif show_ships and self.ship_number != -1:
			print 'S',
		else:
			print ' ',
	
	def has_hit(self):
		return (self.ship_number != -1 and self.has_shot)
	
	def write_xml(self, doc, element):
		element.setAttribute("ship_number", str(self.ship_number))
		element.setAttribute("has_shot", str(int(self.has_shot)))
	
	def read_xml(self, element):
		self.ship_number = int(element.getAttribute("ship_number"))
		self.has_shot = bool(int(element.getAttribute("has_shot")))

class Map:
	"""Contains the ships, shots and the drawable array which is updated synchronously"""

	# These methods are supposed to be used by the user

	def __init__(self, w, h, images):
		"""
		Initialize the map
		
		w, h: width, height
		images: a dictionary of the preloaded images
		"""
		self.w = w
		self.h = h
		self.blocks = []
		for i in range(0,w*h):
			b = Block()
			self.blocks.append(b);
		self.ships = []
		self.shots = []
		self.ship_add_counter = 0
		self.num_ships_destroyed = 0
		self.images = images

	def clear(self):
		"""Clear everything on the map"""
		for i in range(0,self.w*self.h):
			b = Block()
			self.blocks[i] = b
		del self.ships[0:]
		del self.shots[0:]
		self.ship_add_counter = 0
		self.num_ships_destroyed = 0

	def add_ship(self, ship):
		"""Add a ship object on to the map."""
		# Validate ship position
		for (x,y) in ship.get_positions():
			if x < 0 or x >= self.w or y < 0 or y >= self.h:
				raise ValueError('ship position over map border');
			if self.blocks[y*self.w + x].ship_number != -1:
				raise ValueError('can\'t put ship on an another ship');
		# Add the ship
		self.ships.append(ship)
		self.blit_ship(ship, self.ship_add_counter)
		self.ship_add_counter = self.ship_add_counter + 1
	
	def count_free_space(self, x, y, w, h, border_is_free = False):
		"""Count free blocks in the area (those that doesn't contain ships)"""
		free_space = 0
		for j in range(0,h):
			for i in range(0,w):
				if x+i < 0 or y+j < 0 or x+i >= self.w or y+j >= self.h:
					if border_is_free:
						free_space = free_space + 1
					continue
				if self.blocks[(y+j)*self.w + x+i].ship_number == -1:
					free_space = free_space + 1
		return free_space
	
	def add_ships_randomly(self):
		"""Add the ships defined in ship_info on to the map. random seed has to be set before calling."""
		ships_to_add = get_ship_add_list()
		if random.randint(0,666) == 0:
			i = 0
			for name in ships_to_add:
				ship = Ship(self.images)
				ship.set(1+i, 2, False, name)
				self.add_ship(ship)
				i = i + 1
			return
		for name in ships_to_add:
			length = ship_info[name][0]
			horizontal = random.randint(0,1)
			w = 1
			h = 1
			if horizontal:
				w = length
			else:
				h = length
			free_positions = [] # contains tuples ((x,y), free space)
			
			border_is_free = (random.randint(0,1) == 0)
			count_corners = (random.randint(0,2) == 0)

			reverse_j = random.randint(0,1)
			for j_ in range(0, self.h - h + 1):
				if not reverse_j:
					j = j_
				else:
					j = self.h - h - 0 - j_
				reverse_i = random.randint(0,1)
				for i_ in range(0, self.w - w + 1):
					if not reverse_i:
						i = i_
					else:
						i = self.w - w - 0 - i_
					#if there is no space for the ship, don't add it
					if self.count_free_space(i, j, w, h) < length:
						continue;
					if count_corners:
						a = self.count_free_space(i-1, j-1, w+2, h+2, border_is_free = border_is_free)
					else:
						# We ignore the corners and count everything else
						a1 = self.count_free_space(i-1, j, w+2, h, border_is_free = border_is_free)
						a2 = self.count_free_space(i, j-1, w, 1, border_is_free = border_is_free)
						a3 = self.count_free_space(i, j+h, w, 1, border_is_free = border_is_free)
						a = a1 + a2 + a3
					free_positions.append(((i,j), a))
					
			# sort them so that the one having the most space is first
			free_positions_sorted = sorted(free_positions, key = operator.itemgetter(1), reverse=True)
			# get randomly some of the fairly good ones
			#max_id = len(free_positions_sorted) / 4 + 1
			# if there are very few positions, this has to be checked
			#if max_id > len(free_positions_sorted) - 1:
			#	max_id = len(free_positions_sorted) - 1
			#id = random.randint(0, max_id)
			id = 0
			do_it_random = (random.randint(0,4) == 0)
			if do_it_random:
				id = random.randint(0,len(free_positions_sorted)-1)
			x, y = free_positions_sorted[id][0]
			ship = Ship(self.images)
			ship.set(x, y, horizontal, name)
			try:
				self.add_ship(ship)
			except ValueError:
				print "warning: couldn't find place for ship"

	def shoot(self, (x,y)):
		"""Shoot to a position on the map. Return a tuple (bool everything_gone, text message of what happened)"""
		# Validate shot position
		if x < 0 or x >= self.w or y < 0 or y >= self.h:
			raise ValueError('shot position over map border');
		if self.blocks[y*self.w + x].has_shot:
			raise ValueError('shooting to the same position many times doesn\'t make sense')
		# Apply the shot
		self.shots.append((x,y))
		self.blit_shot((x,y))
		# Check what happened
		number = self.blocks[y*self.w + x].ship_number
		if number != -1:
			self.ships[number].add_hit()
			if self.ships[number].destroyed():
				self.num_ships_destroyed = self.num_ships_destroyed + 1
				everything_gone = (self.num_ships_destroyed == self.ship_add_counter)
				return (everything_gone, self.ships[number].name + " destroyed.")
			else:
				return (False, "Something was hit.")
		else:
			return (False, "\"Splash!\"")
	
	def shoot_somewhere(self):
		"""Shoot somewhere on the map, as wisely as possible. Return (everything_gone, position of the shot)"""
		# First, check if there are any hits currently.
		# If there is, try to sink a ship which has been hit already
		for y in range(0,self.h):
			for x in range(0,self.w):
				b = self.blocks[y*self.w + x]
				if b.has_shot and b.ship_number != -1:
					if self.ships[b.ship_number].destroyed():
						continue
					try_blocks = []
					if x != 0:
						try_blocks.append((x-1,y))
					if y != 0:
						try_blocks.append((x,y-1))
					if x != self.w-1:
						try_blocks.append((x+1,y))
					if y != self.h-1:
						try_blocks.append((x,y+1))

					# If there are hits nearby, try finding out whether the
					# ship is vertical or horizontal and shoot accordingly
					
					for i, j in try_blocks:
						c = self.blocks[j*self.w + i]
						if c.has_shot and c.ship_number != -1:
							if self.ships[c.ship_number].destroyed():
								continue
							ship_direction_vector = (i - x, j - y)
							i = x
							j = y
							a = 0
							while a < 4:
								a = a + 1
								(i,j) = (i + ship_direction_vector[0], j + ship_direction_vector[1])
								if i < 0 or j < 0 or i >= self.w or j >= self.h:
									break
								c = self.blocks[j*self.w + i]
								if c.has_shot and c.ship_number == -1:
									break;
								if not c.has_shot:
									everything_gone, message = self.shoot((i, j))
									return (everything_gone, (i,j))
							i = x
							j = y
							a = 0
							while a < 4:
								a = a + 1
								(i,j) = (i - ship_direction_vector[0], j - ship_direction_vector[1])
								if i < 0 or j < 0 or i >= self.w or j >= self.h:
									break
								c = self.blocks[j*self.w + i]
								if c.has_shot and c.ship_number == -1:
									break;
								if not c.has_shot:
									everything_gone, message = self.shoot((i, j))
									return (everything_gone, (i,j))

					# No hits nearby, have to guess the ship's direction
					
					# Take the direction which has the longest continuous
					# space with no misses or destroyed ships
					i = 1
					while x-i >= 0:
						c = self.blocks[(y)*self.w + (x-i)]
						if i == 1 and c.has_shot:
							break;
						if c.has_shot and c.ship_number == -1:
							break
						if c.ship_number != -1:
							if self.ships[c.ship_number].destroyed():
								break
						i = i + 1
					space_left = i-1
					i = 1
					while x+i < self.w:
						c = self.blocks[(y)*self.w + (x+i)]
						if i == 1 and c.has_shot:
							break;
						if c.has_shot and c.ship_number == -1:
							break
						if c.ship_number != -1:
							if self.ships[c.ship_number].destroyed():
								break
						i = i + 1
					space_right = i-1
					i = 1
					while y-i >= 0:
						c = self.blocks[(y-i)*self.w + (x)]
						if i == 1 and c.has_shot:
							break;
						if c.has_shot and c.ship_number == -1:
							break
						if c.ship_number != -1:
							if self.ships[c.ship_number].destroyed():
								break
						i = i + 1
					space_top = i-1
					i = 1
					while y+i < self.h:
						c = self.blocks[(y+i)*self.w + (x)]
						if i == 1 and c.has_shot:
							break;
						if c.has_shot and c.ship_number == -1:
							break
						if c.ship_number != -1:
							if self.ships[c.ship_number].destroyed():
								break
						i = i + 1
					space_bottom = i-1
					
					shoot_position = (0,0)
					if space_left >= space_right and space_left >= space_top and space_left >= space_bottom:
						shoot_position = (x-1,y)
					elif space_right >= space_left and space_right >= space_top and space_right >= space_bottom:
						shoot_position = (x+1,y)
					elif space_top >= space_right and space_top >= space_left and space_top >= space_bottom:
						shoot_position = (x,y-1)
					elif space_bottom >= space_right and space_bottom >= space_top and space_bottom >= space_left:
						shoot_position = (x,y+1)
					everything_gone, message = self.shoot(shoot_position)
					return (everything_gone, shoot_position)

		# Otherwise (if there is no hits already) try to shoot somewhere, using silly heuristics.
		# This looks for positions with the most free space around them.
		unshot_positions = []
		space_wanted = 5+5
		neighbour_hit_ignore = (random.randint(0,3) == 0)
		while space_wanted >= -10 and len(unshot_positions) == 0:
			for y in range(0,self.h):
				for x in range(0,self.w):
					b = self.blocks[y*self.w + x]
					# We are looking for blocks that are not shot already
					if not b.has_shot:
						# If there is a hit in a neighbour block, this isn't a probable location of a ship
						neighbour_hit_count = 0
						if not neighbour_hit_ignore:
							if x-1 >= 0:
								if self.blocks[(y)*self.w + (x-1)].has_hit():
									neighbour_hit_count = neighbour_hit_count + 1
							if x+1 < self.w:
								if self.blocks[(y)*self.w + (x+1)].has_hit():
									neighbour_hit_count = neighbour_hit_count + 1
							if y-1 >= 0:
								if self.blocks[(y-1)*self.w + (x)].has_hit():
									neighbour_hit_count = neighbour_hit_count + 1
							if y+1 < self.h:
								if self.blocks[(y+1)*self.w + (x)].has_hit():
									neighbour_hit_count = neighbour_hit_count + 1
						# If there are not enough unshot blocks near this one, this isn't a good target
						i = 1
						while x-i >= 0:
							c = self.blocks[(y)*self.w + (x-i)]
							if c.has_shot:
								break
							i = i + 1
						space_left = i-1
						i = 1
						while x+i < self.w:
							c = self.blocks[(y)*self.w + (x+i)]
							if c.has_shot:
								break
							i = i + 1
						space_right = i-1
						i = 1
						while y-i >= 0:
							c = self.blocks[(y-i)*self.w + (x)]
							if c.has_shot:
								break
							i = i + 1
						space_top = i-1
						i = 1
						while y+i < self.h:
							c = self.blocks[(y+i)*self.w + (x)]
							if c.has_shot:
								break
							i = i + 1
						space_bottom = i-1

						if space_left > 4:
							space_left = 4
						if space_right > 4:
							space_right = 4
						if space_top > 4:
							space_top = 4
						if space_bottom > 4:
							space_bottom = 4

						space_avg = (space_left + space_right + space_top + space_bottom) / 4

						if space_left <= space_right and space_left <= space_top and space_left <= space_bottom:
							space_min = space_left
						if space_right <= space_left and space_right <= space_top and space_right <= space_bottom:
							space_min = space_right
						if space_top <= space_right and space_top <= space_left and space_top <= space_bottom:
							space_min = space_top
						if space_bottom <= space_right and space_bottom <= space_top and space_bottom <= space_top:
							space_min = space_bottom

						if space_min + space_avg - neighbour_hit_count >= space_wanted:
							unshot_positions.append((x,y))
							
			# If nothing was found, go and retry with a smaller wanted amount of space
			space_wanted = space_wanted - 1
							
		if len(unshot_positions) != 0:
			# And of course, random!
			# We're trying to act as clever as possible, anyway 8)
			n = random.randint(0, len(unshot_positions)-1)
			everything_gone, message = self.shoot(unshot_positions[n])
			return (everything_gone, unshot_positions[n])
		
		# Well, this can't happen anyway...
		return (False, (-1,-1))


	def print_(self, show_ships=True):
		"""Print the map to stdout"""
		print '   ',
		for i in range(0, self.w):
			print chr(65+i),
		print ''
		for j in range(0, self.h):
			print j,
			print '|',
			for i in range(0, self.w):
				self.blocks[j*self.w + i].print_(show_ships)
			print '|'
	
	def draw(self, canvas, show_ships, x=0, y=0):
		"""Draw the map on a Tkinter.Canvas"""
		global images
		for ship in self.ships:
			if show_ships or ship.destroyed():
				ship.draw(canvas, x, y)
		for j in range(0,self.h):
			for i in range(0,self.w):
				b = self.blocks[j*self.w + i]
				if b.has_shot:
					if b.ship_number != -1:
						if self.ships[b.ship_number].destroyed():
							# The ship was drawn already
							pass
						else:
							canvas.create_image(x+i*BLOCK_W, y+j*BLOCK_H, image=self.images['explosion'], anchor=Tkinter.NW)
					else:
						canvas.create_image(x+i*BLOCK_W, y+j*BLOCK_H, image=self.images['splash'], anchor=Tkinter.NW)
				else:
					if b.ship_number != -1 and show_ships:
						# The ship was drawn already
						pass
					elif show_ships:
						canvas.create_image(x+i*BLOCK_W, y+j*BLOCK_H, image=self.images['water'], anchor=Tkinter.NW)
					else:
						canvas.create_image(x+i*BLOCK_W, y+j*BLOCK_H, image=self.images['waterunknown'], anchor=Tkinter.NW)
						
	def write_xml(self, doc, element):
		"""Add the save data to an xml element"""
		element.setAttribute("num_ships_destroyed", str(self.num_ships_destroyed))
		ships = doc.createElement("ships")
		element.appendChild(ships)
		for ship in self.ships:
			ship_elem = doc.createElement("ship")
			ships.appendChild(ship_elem)
			ship.write_xml(doc, ship_elem)
		blocks = doc.createElement("blocks")
		element.appendChild(blocks)
		for i in range(0,self.w*self.h):
			b = self.blocks[i]
			block_elem = doc.createElement("block")
			blocks.appendChild(block_elem)
			b.write_xml(doc, block_elem)
	
	def read_xml(self, element):
		"""Read the save data from an xml element"""
		self.clear()

		self.num_ships_destroyed = int(element.getAttribute("num_ships_destroyed"))
		ships = element.getElementsByTagName("ships")[0]
		for node in ships.getElementsByTagName("ship"):
			ship = Ship(self.images)
			ship.read_xml(node)
			self.ships.append(ship)
			self.ship_add_counter = self.ship_add_counter + 1

		blocks = element.getElementsByTagName("blocks")[0]
		nodes = blocks.getElementsByTagName("block")
		if len(nodes) != self.w*self.h:
			print "warning: wrong number of blocks in xml"
		for i in range(0,self.w*self.h):
			self.blocks[i].read_xml(nodes[i])
			
	# These are 'private'

	def blit_ship(self, ship, number):
		for (x,y) in ship.get_positions():
			self.blocks[y*self.w + x].ship_number = number
	
	def blit_shot(self, (x,y)):
		self.blocks[y*self.w + x].has_shot = True
	

